(* sequence helper function *)
structure SeqUtils =
struct

    (* determines wehther there exists an element of a sequence satisfying
       the predicate p (see lab 8)
     *)
    fun seqExists(p : 'a -> bool, s : 'a Seq.seq) : bool =
        Seq.mapreduce(p, false, fn (x,y) => x orelse y, s)

end

structure Player =
struct

    datatype player = X | O

    fun player_to_string (p : player) : string =
        case p of
            X => "X"
          | O => "O"

    fun next (p : player) : player =
        case p of
            X => O
          | O => X

    datatype status = Playing | Won of player | Draw 
        
end

signature GAME =
sig

    type board

    val show_board : board -> string

    type state = board * Player.player 
        
    val start : state
        
    val hash : state -> string

    val check_status : state -> Player.status

    type move

    val parse_move : state * string -> move option
    val show_move : move -> string

    val possible_moves : state -> move Seq.seq

    val make_move : state * move -> state
        
end

structure CountDown :> GAME =
struct

    (* this means you can write the variables bound in the Player module without writing Player.variable,
       e.g. X instead of Player.X *)
    open Player
    
    type board = int 

    type move = int

    type state = board * player

    val start = (21, X)

    fun show_board n = Int.toString n ^ " pieces left "

    fun hash (b,who) = Int.toString b ^ player_to_string who

    fun check_status (b,p) =
        case b of
            0 => Won (next p)
          | _ => Playing

    fun possible_moves (n,_) = Seq.tabulate(fn i => i+1 , Int.min(n, 3))

    fun make_move ((b, who), i) = (b - i , next who)

    fun parse_move((b,who), s) =
        case Int.fromString s of
            SOME i =>
                (case SeqUtils.seqExists(fn j => j=i, possible_moves (b,who)) of
                     true => SOME i
                   | false => NONE)
          | NONE => NONE

    fun show_move i = Int.toString i

end

structure Game :> GAME = CountDown

signature TRAIN =
sig

    type memory = (string,real) Dict.dict

    val best_next_state : memory * Game.state -> Game.state * real
        
    val train : int * memory * Game.state -> memory

end 

structure Train :> TRAIN =
struct

    open Player

    local
        val r = Random.rand (05052023, 33409756234)
    in
        (* assuming low <= hi, generate a random integer 
           in the range [low...hi] including both low and hi *)
        fun randRange (low : int,hi : int) : int =
            case low = hi of
                true => low
              | false => Random.randRange(low,hi) r 
    end

    type memory = (string,real) Dict.dict

    val update_rate = 0.3
        
    fun best_next_state(mem : memory, (board, who) : Game.state) : Game.state * real =
        let val (move) = Seq.nth(0,Game.possible_moves (board,who)) in 
            (Game.make_move((board,who),move), 0.0)
        end 

    fun train(n : int, mem : memory, (board,who) : Game.state) = mem
        
end 

structure Controller =
struct

    (* this means you can write the variables bound in the Player module without writing Player.variable,
       e.g. X instead of Player.X *)
    open Player 
    
    val mem = Train.train(100000, Dict.empty, Game.start)

    fun controller (s : Game.state) : unit = ()

    fun go() = controller(Game.start)

end



